/* This file is part of Legendes II */
/* Copyright (c) 1998-1999 by
 *   Antoine Allain (allain@aix.ensam.fr)
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   Benjamin Lerman (benjamin.lerman@ens.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File handler.c */

/* Thus speaketh ye mighty CVS:
 * $Id: handler.c,v 1.12 1999/05/30 22:23:03 david Exp $ */

/* This program contains the main function of the handler threads. */

/* CVS identity of this file. */
static char id[] = "$Id: handler.c,v 1.12 1999/05/30 22:23:03 david Exp $";

/* Include files */

#ifndef _THREAD_SAFE
#define _THREAD_SAFE		/* Make sure libc is reentrant. */
#endif

#include <stdio.h>		/* Standard I/O */
#include <stdlib.h>		/* Standard library functions */

#include <unistd.h>		/* UNIX standard functions */

#include <pthread.h>		/* Posix threads */

#include <syslog.h>		/* For priority values */


#include "server.h"

#include "lgdproto.h"

#include "map.h"

/* Main functions */
void
send_reply (struct lsock_s *sck, union lgd_proto_msg_u *rpy)
     /* Send a reply. */
{
  if ( ! sck->sopen ) return;
  if ( write (sck->s, rpy, ntohl(rpy->f.len)) < 0 )
    {
      /* Just close the socket and continue. */
      close (sck->s);
      sck->sopen = 0;
      return;
    }
}

void
reply_error (struct lsock_s *sck, uint32_t callback,
	     uint32_t errcode)
     /* Return an error code. */
{
  union lgd_proto_msg_u *rpy;

  rpy = alloca (20);
  rpy->f.code = htonl (LGD_PROTO_REP_ERR);
  rpy->f.len = htonl (20);
  rpy->f.res = htonl (0); /* Reserved */
  rpy->f.callback = callback;
  rpy->f.data[0] = htonl (errcode);
  send_reply (sck, rpy);
}

void
reply_nothing (struct lsock_s *sck, uint32_t callback)
     /* Make a no-content response. */
{
  union lgd_proto_msg_u *rpy;

  rpy = alloca (16);
  rpy->f.code = htonl (LGD_PROTO_REP_SYN);
  rpy->f.len = htonl (16);
  rpy->f.res = htonl (0); /* Reserved */
  rpy->f.callback = callback;
  send_reply (sck, rpy);
}

void
reply_simple (struct lsock_s *sck, uint32_t callback,
	      uint32_t reply)
     /* Return just a 32-bit quantity. */
{
  union lgd_proto_msg_u *rpy;

  rpy = alloca (20);
  rpy->f.code = htonl (LGD_PROTO_REP_SYN);
  rpy->f.len = htonl (20);
  rpy->f.res = htonl (0); /* Reserved */
  rpy->f.callback = callback;
  rpy->f.data[0] = htonl (reply);
  send_reply (sck, rpy);
}

void
reply_double (struct lsock_s *sck, uint32_t callback,
	      uint32_t reply1, uint32_t reply2)
     /* Return two 32-bit quantities. */
{
  union lgd_proto_msg_u *rpy;

  rpy = alloca (24);
  rpy->f.code = htonl (LGD_PROTO_REP_SYN);
  rpy->f.len = htonl (24);
  rpy->f.res = htonl (0); /* Reserved */
  rpy->f.callback = callback;
  rpy->f.data[0] = htonl (reply1);
  rpy->f.data[1] = htonl (reply2);
  send_reply (sck, rpy);
}

void
do_greet (struct lsock_s *sck)
{
  union lgd_proto_msg_u *rpy;

  rpy = alloca (24);
  rpy->f.code = htonl (LGD_PROTO_REP_CON);
  rpy->f.len = htonl (24);
  rpy->f.res = htonl (0); /* Reserved */
  rpy->f.callback = htonl (LGD_PROTO_MAGIC);
  rpy->f.data[0] = htonl (LGD_PROTO_VERNUM);
  rpy->f.data[1] = htonl (1);
  send_reply (sck, rpy);
  sck->state = CONN_STATE_NORMAL;
}

void
do_request (struct lsock_s *sck)
     /* Handle a request on the given connection. */
{
  uint32_t len,req,callback;
  if ( sck->buflen < 8 )
    /* We don't have the request code and length yet. */
    {
      if ( ! sck->sopen )
	/* Socket is closed so new data isn't likely to arrive. */
	{
	  sck->state = CONN_STATE_DEAD;
	  sck->buflen = 0;
	  sck->handwork = 1; /* Loop again so as to close connection */
	  return;
	}
      else
	sck->handwork = 0; /* We've done all we could for the moment */
      return;
    }
  req = ntohl (((union lgd_proto_msg_u *)(sck->buffer))->f.code);
  len = ntohl (((union lgd_proto_msg_u *)(sck->buffer))->f.len);
  if ( len < 8 )
    /* Length is too small to be reasonable.  There is nothing to do
     * but complain and then close the socket: we are out of sync. */
    {
      reply_error (sck, htonl (0), LGD_PROTO_ERR_TOOLONG);
      /* XXX - TOOLONG?  That doesn't make sense... */
      sck->state = CONN_STATE_DEAD;
      sck->buflen = 0;
      sck->handwork = 1; /* Loop again so as to close connection */
      return;
    }
  if ( ( len >= 16 ) && ( sck->buflen < 16 ) && sck->sopen )
    /* A callback is coming but we don't have it yet.  Wait for it. */
    {
      sck->handwork = 0;
      return;
    }
  else if ( sck->buflen >= 16 )
    callback = ((union lgd_proto_msg_u *)(sck->buffer))->f.callback;
  else
    /* Request is too small to contain a callback.  Assume 0. */
    callback = htonl (0); /* Yes, this is htonl() and not ntohl(). */
  if ( len > BUFSIZE )
    /* Request is too long to fit in buffer no matter what.
     * Complain. */
    {
      serverlog (LOG_WARNING, "Request too long (len=%d)", len);
      reply_error (sck, callback, LGD_PROTO_ERR_TOOLONG);
      sck->state = CONN_STATE_DEAD;
      sck->buflen = 0;
      sck->handwork = 1; /* Loop again so as to close connection */
      return;
    }
  if ( sck->buflen < len )
    /* We don't have the full request yet. */
    {
      if ( ! sck->sopen )
	/* Socket is closed so new data isn't likely to arrive. */
	/* Proceed with what we have. */
	len = sck->buflen;
      else
	/* Wait for more data */
	{
	  sck->handwork = 0;
	  return;
	}
    }
  serverlog (LOG_DEBUG, "Treating request: req=%d, len=%d", req, len);
  switch ( req )
    {
    case LGD_PROTO_REQ_NOP:
      /* NOP */
      reply_nothing (sck, callback);
      break;
    case LGD_PROTO_REQ_ACK:
      /* ACK? */
      reply_simple (sck, callback, 1);
      break;
    case LGD_PROTO_REQ_MAGIC:
      /* MAGIC */
      reply_simple (sck, callback, LGD_PROTO_MAGIC);
      break;
    case LGD_PROTO_REQ_VERNUM:
      /* VERNUM */
      reply_simple (sck, callback, LGD_PROTO_VERNUM);
      break;
    case LGD_PROTO_REQ_DIE:
      /* DIE */
      reply_nothing (sck, callback);
      exit (1);
      break;
    case LGD_PROTO_REQ_IDENT:
      /* IDENT */
      if ( ( len >= 16 )
	   && ( strncmp (sck->buffer+16, "legendes", len-16) == 0 ) )
	reply_simple (sck, callback, 1);
      else
	reply_error (sck, callback, LGD_PROTO_ERR_REFUSED);
      break;
    case LGD_PROTO_REQ_MAPQ:
      /* MAPQ */
      {
	unsigned long x, y;
	if ( len >= 20 )
	  x = ntohl (((union lgd_proto_msg_u *)(sck->buffer))->f.data[0]);
	else
	  x = 0;
	if ( len >= 24 )
	  y = ntohl (((union lgd_proto_msg_u *)(sck->buffer))->f.data[1]);
	else
	  y = 0;
	if ( ( x >= 0 ) && ( y >= 0 )
	     && ( x < MAP_SIZE_X ) && ( y < MAP_SIZE_Y ) )
	  reply_simple (sck, callback, read_map (x, y));
	else
	  reply_error (sck, callback, LGD_PROTO_ERR_OBOUNDS);
      }
      break;
    case LGD_PROTO_REQ_WHERE:
      /* WHERE */
      reply_double (sck, callback, sck->x, sck->y);
      break;
    default:
      reply_error (sck, callback, LGD_PROTO_ERR_UNKNOWN);
      break;
    }
  memmove (sck->buffer, sck->buffer+len, len);
  sck->buflen -= len;
  /* *Don't* set sck->handwork back to 0, because there may be another
   * request in the buffer. */
}

void
handler_loop (struct lsock_s *sck)
     /* This function performs a loop that handles the requests that
      * the listener thread has put in the connection buffer. */
{
  safe_pthread_mutex_lock (&sck->d_mutex);
  while (1)
    {
      while ( ! sck->handwork )
	pthread_cond_wait (&sck->h_cond, &sck->d_mutex);
      /* Do everything */
      switch ( sck->state )
	{
	case CONN_STATE_NEW:
	  do_greet (sck);
	  break;
	case CONN_STATE_NORMAL:
	  do_request (sck);
	  break;
	case CONN_STATE_DEAD:
	  if ( sck->sopen )
	    {
	      close (sck->s);
	      sck->sopen = 0;
	    }
	  /* The time has come to die now */
	  safe_pthread_mutex_unlock (&sck->d_mutex);
	  pthread_exit (NULL);
	  break;
	}
    }
}
