/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   Benjamin Lerman (benjamin.lerman@ens.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File server.c */

/* Thus speaketh ye mighty CVS:
 * $Id: server.c,v 1.31 1999/05/30 22:23:03 david Exp $ */

/* This program opens a network socket (on port PORTNUM) and listens
 * to it.  That's about all.  It's still pretty complicated because
 * it's threaded.  We have a first bunch of threads, called listener
 * threads (LISTENER_THREADS_NB of them), which take care of reading
 * whatever arrives on the socket and copying it to data buffers
 * associated to each connection.  A second bunch of threads, handler
 * threads (HANDLER_THREADS_NB of them) reads these buffers and
 * interprets them.  The complication comes from the fact that within
 * each of these groups the threads are undifferentiated: there isn't
 * one listener thread per connection or some such thing, but rather,
 * each thread takes care of all connections at once - so we have to
 * set up a lot of mutexes and complicated rules to stop them from
 * stepping on each others' toes. */

/* CVS identity of this file. */
static char id[] = "$Id: server.c,v 1.31 1999/05/30 22:23:03 david Exp $";

/* Include files */

#ifndef _THREAD_SAFE
#define _THREAD_SAFE		/* Make sure libc is reentrant. */
#endif

#include <stdio.h>		/* Standard I/O */
#include <stdlib.h>		/* Standard library functions */

#include <errno.h>		/* Error numbers */
#include <string.h>		/* String functions such as strerror() */

#include <sys/types.h>		/* System types */
#include <sys/socket.h>		/* Socket types */
#include <sys/time.h>		/* For struct timeval */
#include <unistd.h>		/* UNIX standard functions */
#include <fcntl.h>		/* For fcntl() and its constants */

#include <netinet/in.h>		/* INET domain (TCP/IP) sockets */

#include <netdb.h>		/* gethostbyaddr() and friends */

#include <pthread.h>		/* Posix threads */

#include <syslog.h>

#include <stdarg.h>


#include "server.h"

#include "map.h"

/* Main functions */

void
serverlog (int priority, char *fmt, ...)
{
#if USE_SYSLOG
  static int initialized = 0;
#endif
  va_list ap;

  if ( priority > LOG_LEVEL )
    return;
#if USE_SYSLOG
  if ( ! initialized )
    {
      openlog ("leglab", LOG_PID, LOG_USER);
      initialized = 1;
    }
#endif
  va_start (ap, fmt);
#if USE_SYSLOG
  vsyslog (priority, fmt, ap);
#else
  fprintf (stderr, "<%d>[%05d]", priority, getpid ());
  vfprintf (stderr, fmt, ap);
  fprintf (stderr, "\n");
#endif
  va_end (ap);
}

void *
entry_accepter (void *garbage)
     /* Entry point for the accepter thread */
{
  accepter_loop ();
  return NULL;
}

int
main(void)
     /* Abandon all hope, ye who enter here */
{
  pthread_t garbage;

  /* Initialize global socket */
  init_gsock();
  /* Initialize map subunit */
  init_map();
  /* Create accepter thread */
  pthread_create (&garbage, NULL, entry_accepter, NULL);
  /* The father thread dies here */
  pthread_exit(NULL);
  return 0; /* Keep gcc happy */
}
